'use strict';

var _require = require('./lib/task');

var Task = _require.Task;

var _require2 = require('./lib/taskgroup');

var TaskGroup = _require2.TaskGroup;

module.exports = { Task: Task, TaskGroup: TaskGroup };